"""Generated wrapper for IRoyalty Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for IRoyalty below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        IRoyaltyValidator,
    )
except ImportError:

    class IRoyaltyValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""


try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass





class GetDefaultRoyaltyInfoMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the getDefaultRoyaltyInfo method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address)
        self._underlying_method = contract_function

    def call(self, tx_params: Optional[TxParams] = None) -> Tuple[str, int]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method().call(tx_params.as_dict())
        return (returned[0],returned[1],)

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())

class GetRoyaltyInfoForTokenMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the getRoyaltyInfoForToken method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, token_id: int):
        """Validate the inputs to the getRoyaltyInfoForToken method."""
        self.validator.assert_valid(
            method_name='getRoyaltyInfoForToken',
            parameter_name='tokenId',
            argument_value=token_id,
        )
        # safeguard against fractional inputs
        token_id = int(token_id)
        return (token_id)

    def call(self, token_id: int, tx_params: Optional[TxParams] = None) -> Tuple[str, int]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (token_id) = self.validate_and_normalize_inputs(token_id)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(token_id).call(tx_params.as_dict())
        return (returned[0],returned[1],)

    def send_transaction(self, token_id: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (token_id) = self.validate_and_normalize_inputs(token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(token_id).transact(tx_params.as_dict())

    def build_transaction(self, token_id: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (token_id) = self.validate_and_normalize_inputs(token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(token_id).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, token_id: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (token_id) = self.validate_and_normalize_inputs(token_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(token_id).estimateGas(tx_params.as_dict())

class RoyaltyInfoMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the royaltyInfo method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, token_id: int, sale_price: int):
        """Validate the inputs to the royaltyInfo method."""
        self.validator.assert_valid(
            method_name='royaltyInfo',
            parameter_name='tokenId',
            argument_value=token_id,
        )
        # safeguard against fractional inputs
        token_id = int(token_id)
        self.validator.assert_valid(
            method_name='royaltyInfo',
            parameter_name='salePrice',
            argument_value=sale_price,
        )
        # safeguard against fractional inputs
        sale_price = int(sale_price)
        return (token_id, sale_price)

    def call(self, token_id: int, sale_price: int, tx_params: Optional[TxParams] = None) -> Tuple[str, int]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (token_id, sale_price) = self.validate_and_normalize_inputs(token_id, sale_price)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(token_id, sale_price).call(tx_params.as_dict())
        return (returned[0],returned[1],)

    def send_transaction(self, token_id: int, sale_price: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (token_id, sale_price) = self.validate_and_normalize_inputs(token_id, sale_price)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(token_id, sale_price).transact(tx_params.as_dict())

    def build_transaction(self, token_id: int, sale_price: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (token_id, sale_price) = self.validate_and_normalize_inputs(token_id, sale_price)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(token_id, sale_price).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, token_id: int, sale_price: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (token_id, sale_price) = self.validate_and_normalize_inputs(token_id, sale_price)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(token_id, sale_price).estimateGas(tx_params.as_dict())

class SetDefaultRoyaltyInfoMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the setDefaultRoyaltyInfo method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, royalty_recipient: str, royalty_bps: int):
        """Validate the inputs to the setDefaultRoyaltyInfo method."""
        self.validator.assert_valid(
            method_name='setDefaultRoyaltyInfo',
            parameter_name='_royaltyRecipient',
            argument_value=royalty_recipient,
        )
        royalty_recipient = self.validate_and_checksum_address(royalty_recipient)
        self.validator.assert_valid(
            method_name='setDefaultRoyaltyInfo',
            parameter_name='_royaltyBps',
            argument_value=royalty_bps,
        )
        # safeguard against fractional inputs
        royalty_bps = int(royalty_bps)
        return (royalty_recipient, royalty_bps)

    def call(self, royalty_recipient: str, royalty_bps: int, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (royalty_recipient, royalty_bps) = self.validate_and_normalize_inputs(royalty_recipient, royalty_bps)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(royalty_recipient, royalty_bps).call(tx_params.as_dict())

    def send_transaction(self, royalty_recipient: str, royalty_bps: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (royalty_recipient, royalty_bps) = self.validate_and_normalize_inputs(royalty_recipient, royalty_bps)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(royalty_recipient, royalty_bps).transact(tx_params.as_dict())

    def build_transaction(self, royalty_recipient: str, royalty_bps: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (royalty_recipient, royalty_bps) = self.validate_and_normalize_inputs(royalty_recipient, royalty_bps)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(royalty_recipient, royalty_bps).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, royalty_recipient: str, royalty_bps: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (royalty_recipient, royalty_bps) = self.validate_and_normalize_inputs(royalty_recipient, royalty_bps)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(royalty_recipient, royalty_bps).estimateGas(tx_params.as_dict())

class SetRoyaltyInfoForTokenMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the setRoyaltyInfoForToken method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, token_id: int, recipient: str, bps: int):
        """Validate the inputs to the setRoyaltyInfoForToken method."""
        self.validator.assert_valid(
            method_name='setRoyaltyInfoForToken',
            parameter_name='tokenId',
            argument_value=token_id,
        )
        # safeguard against fractional inputs
        token_id = int(token_id)
        self.validator.assert_valid(
            method_name='setRoyaltyInfoForToken',
            parameter_name='recipient',
            argument_value=recipient,
        )
        recipient = self.validate_and_checksum_address(recipient)
        self.validator.assert_valid(
            method_name='setRoyaltyInfoForToken',
            parameter_name='bps',
            argument_value=bps,
        )
        # safeguard against fractional inputs
        bps = int(bps)
        return (token_id, recipient, bps)

    def call(self, token_id: int, recipient: str, bps: int, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (token_id, recipient, bps) = self.validate_and_normalize_inputs(token_id, recipient, bps)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(token_id, recipient, bps).call(tx_params.as_dict())

    def send_transaction(self, token_id: int, recipient: str, bps: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (token_id, recipient, bps) = self.validate_and_normalize_inputs(token_id, recipient, bps)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(token_id, recipient, bps).transact(tx_params.as_dict())

    def build_transaction(self, token_id: int, recipient: str, bps: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (token_id, recipient, bps) = self.validate_and_normalize_inputs(token_id, recipient, bps)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(token_id, recipient, bps).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, token_id: int, recipient: str, bps: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (token_id, recipient, bps) = self.validate_and_normalize_inputs(token_id, recipient, bps)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(token_id, recipient, bps).estimateGas(tx_params.as_dict())

# pylint: disable=too-many-public-methods,too-many-instance-attributes
class IRoyalty:
    """Wrapper class for IRoyalty Solidity contract."""
    get_default_royalty_info: GetDefaultRoyaltyInfoMethod
    """Constructor-initialized instance of
    :class:`GetDefaultRoyaltyInfoMethod`.
    """

    get_royalty_info_for_token: GetRoyaltyInfoForTokenMethod
    """Constructor-initialized instance of
    :class:`GetRoyaltyInfoForTokenMethod`.
    """

    royalty_info: RoyaltyInfoMethod
    """Constructor-initialized instance of
    :class:`RoyaltyInfoMethod`.
    """

    set_default_royalty_info: SetDefaultRoyaltyInfoMethod
    """Constructor-initialized instance of
    :class:`SetDefaultRoyaltyInfoMethod`.
    """

    set_royalty_info_for_token: SetRoyaltyInfoForTokenMethod
    """Constructor-initialized instance of
    :class:`SetRoyaltyInfoForTokenMethod`.
    """


    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        validator: IRoyaltyValidator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param web3_or_provider: Either an instance of `web3.Web3`:code: or
            `web3.providers.base.BaseProvider`:code:
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        # pylint: disable=too-many-statements

        self.contract_address = contract_address

        if not validator:
            validator = IRoyaltyValidator(web3_or_provider, contract_address)

        web3 = None
        if isinstance(web3_or_provider, BaseProvider):
            web3 = Web3(web3_or_provider)
        elif isinstance(web3_or_provider, Web3):
            web3 = web3_or_provider
        else:
            raise TypeError(
                "Expected parameter 'web3_or_provider' to be an instance of either"
                + " Web3 or BaseProvider"
            )

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                         middleware['function'], layer=middleware['layer'],
                    )
            except ValueError as value_error:
                if value_error.args == ("You can't add the same un-named instance twice",):
                    pass

        self._web3_eth = web3.eth

        functions = self._web3_eth.contract(address=to_checksum_address(contract_address), abi=IRoyalty.abi()).functions

        self.get_default_royalty_info = GetDefaultRoyaltyInfoMethod(web3_or_provider, contract_address, functions.getDefaultRoyaltyInfo)

        self.get_royalty_info_for_token = GetRoyaltyInfoForTokenMethod(web3_or_provider, contract_address, functions.getRoyaltyInfoForToken, validator)

        self.royalty_info = RoyaltyInfoMethod(web3_or_provider, contract_address, functions.royaltyInfo, validator)

        self.set_default_royalty_info = SetDefaultRoyaltyInfoMethod(web3_or_provider, contract_address, functions.setDefaultRoyaltyInfo, validator)

        self.set_royalty_info_for_token = SetRoyaltyInfoForTokenMethod(web3_or_provider, contract_address, functions.setRoyaltyInfoForToken, validator)

    def get_default_royalty_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for DefaultRoyalty event.

        :param tx_hash: hash of transaction emitting DefaultRoyalty event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=IRoyalty.abi()).events.DefaultRoyalty().processReceipt(tx_receipt)
    def get_royalty_for_token_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for RoyaltyForToken event.

        :param tx_hash: hash of transaction emitting RoyaltyForToken event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=IRoyalty.abi()).events.RoyaltyForToken().processReceipt(tx_receipt)

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"anonymous":false,"inputs":[{"indexed":false,"internalType":"address","name":"newRoyaltyRecipient","type":"address"},{"indexed":false,"internalType":"uint256","name":"newRoyaltyBps","type":"uint256"}],"name":"DefaultRoyalty","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"uint256","name":"tokenId","type":"uint256"},{"indexed":false,"internalType":"address","name":"royaltyRecipient","type":"address"},{"indexed":false,"internalType":"uint256","name":"royaltyBps","type":"uint256"}],"name":"RoyaltyForToken","type":"event"},{"inputs":[],"name":"getDefaultRoyaltyInfo","outputs":[{"internalType":"address","name":"","type":"address"},{"internalType":"uint16","name":"","type":"uint16"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"uint256","name":"tokenId","type":"uint256"}],"name":"getRoyaltyInfoForToken","outputs":[{"internalType":"address","name":"","type":"address"},{"internalType":"uint16","name":"","type":"uint16"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"uint256","name":"tokenId","type":"uint256"},{"internalType":"uint256","name":"salePrice","type":"uint256"}],"name":"royaltyInfo","outputs":[{"internalType":"address","name":"receiver","type":"address"},{"internalType":"uint256","name":"royaltyAmount","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"_royaltyRecipient","type":"address"},{"internalType":"uint256","name":"_royaltyBps","type":"uint256"}],"name":"setDefaultRoyaltyInfo","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"uint256","name":"tokenId","type":"uint256"},{"internalType":"address","name":"recipient","type":"address"},{"internalType":"uint256","name":"bps","type":"uint256"}],"name":"setRoyaltyInfoForToken","outputs":[],"stateMutability":"nonpayable","type":"function"}]'  # noqa: E501 (line-too-long)
        )

# pylint: disable=too-many-lines



from re import M
from typing import Any, Dict, Optional, cast

from eth_utils import is_address
from thirdweb.abi import Marketplace
from thirdweb.abi import IERC165
from thirdweb.abi.ierc1155 import IERC1155
from thirdweb.abi.ierc721 import IERC721
from thirdweb.abi.marketplace import IMarketplaceListingParameters
from thirdweb.common.currency import (
    fetch_currency_value,
    is_native_token,
    normalize_price_value,
    set_erc20_allowance,
)
from web3.constants import MAX_INT
from thirdweb.common.error import ListingNotFoundException, WrongListingTypeException
from thirdweb.common.marketplace import (
    handle_token_approval,
    is_token_approved_for_marketplace,
    map_offer,
    validate_new_listing_param,
)
from thirdweb.common.nft import fetch_token_metadata_for_contract
from thirdweb.constants.contract import INTERFACE_ID_IERC1155, INTERFACE_ID_IERC721
from thirdweb.constants.currency import ZERO_ADDRESS
from thirdweb.core.classes.base_contract import BaseContract
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.types.currency import Price
from thirdweb.types.marketplace import (
    ContractListing,
    ContractOffer,
    DirectListing,
    ListingType,
    NewDirectListing,
    Offer,
)
from web3.eth import TxReceipt


class MarketplaceDirect(BaseContract[Marketplace]):
    _storage: IpfsStorage

    def __init__(self, contract_wrapper: ContractWrapper, storage: IpfsStorage):
        super().__init__(contract_wrapper)
        self._storage = storage

    """
    READ FUNCTIONS
    """

    def get_listing(self, listing_id: int) -> DirectListing:
        """
        Get a direct listing from the marketplace by ID

        :param listing_id: The ID of the listing to get
        :return: The listing
        """
        raw_listing = self._contract_wrapper._contract_abi.listings.call(listing_id)
        listing = ContractListing(*raw_listing)

        if listing.asset_contract == ZERO_ADDRESS:
            raise ListingNotFoundException(listing_id)

        if ListingType(listing.listing_type) != ListingType.DIRECT:
            raise WrongListingTypeException(
                listing_id,
                "Auction",
                "Direct",
            )

        return self._map_listing(listing)

    def get_active_offer(self, listing_id: int, address: str) -> Optional[Offer]:
        """
        Get an active offer for a direct listing

        :param listing_id: The ID of the listing to get the offer for
        :param address: The address of the user to get the offer for
        :return: The offer
        """
        self._validate_listing(listing_id)

        if not is_address(address):
            raise Exception("Address must be a valid address")

        raw_ofers = self._contract_wrapper._contract_abi.offers.call(
            listing_id, address
        )
        offers = ContractOffer(*raw_ofers)

        if offers.offeror == ZERO_ADDRESS:
            return None

        return map_offer(self._contract_wrapper.get_provider(), listing_id, offers)

    """
    WRITE FUNCTIONS
    """

    def create_listing(self, listing: NewDirectListing) -> int:
        """
        Create a new direct listing

        :param listing: The listing to create
        :return: The ID of the listing
        """
        validate_new_listing_param(listing)

        handle_token_approval(
            self._contract_wrapper.get_provider(),
            self._contract_wrapper.get_signer(),
            self.get_address(),
            listing.asset_contract_address,
            listing.token_id,
            self._contract_wrapper.get_signer_address(),
        )

        normalized_price_per_token = normalize_price_value(
            self._contract_wrapper.get_provider(),
            listing.buyout_price_per_token,
            listing.currency_contract_address,
        )

        receipt = self._contract_wrapper.send_transaction(
            "create_listing",
            [
                IMarketplaceListingParameters(
                    assetContract=listing.asset_contract_address,
                    tokenId=listing.token_id,
                    startTime=listing.start_time_in_seconds,
                    secondsUntilEndTime=listing.listing_duration_in_seconds,
                    quantityToList=listing.quantity,
                    currencyToAccept=listing.currency_contract_address,
                    reservePricePerToken=normalized_price_per_token,
                    buyoutPricePerToken=normalized_price_per_token,
                    listingType=0,
                ),
            ],
        )

        events = self._contract_wrapper.get_events("ListingAdded", receipt)
        return cast(Any, events[0].get("args")).get("listingId")

    def make_offer(
        self,
        listing_id: int,
        quantity_desired: int,
        currency_contract_address: str,
        price_per_token: Price,
        expiration_date: int = int(MAX_INT, 0),
    ) -> TxReceipt:
        """
        Make an offer on a direct listing

        :param listing_id: The ID of the listing to make the offer on
        :param quantity_desired: The quantity desired
        :param currency_contract_address: The address of the currency contract
        :param price_per_token: The price per token
        :return: The transaction receipt
        """
        if is_native_token(currency_contract_address):
            raise Exception(
                "You must used the wrapped native token address when making an offer"
            )

        normalized_price = normalize_price_value(
            self._contract_wrapper.get_provider(),
            price_per_token,
            currency_contract_address,
        )

        try:
            self.get_listing(listing_id)
        except:
            raise ListingNotFoundException(listing_id)

        overrides: Dict[Any, Any] = {}
        set_erc20_allowance(
            self._contract_wrapper,
            quantity_desired * normalized_price,
            currency_contract_address,
            overrides,
        )

        print(
            "PARAMS: ",
            [
                listing_id,
                quantity_desired,
                currency_contract_address,
                normalized_price,
                expiration_date,
            ],
        )

        # TODO: Add OVERRIDES
        return self._contract_wrapper.send_transaction(
            "offer",
            [
                listing_id,
                quantity_desired,
                currency_contract_address,
                normalized_price,
                expiration_date,
            ],
        )

    def accept_offer(self, listing_id: int, address_or_offerror: str) -> TxReceipt:
        """
        Accept a direct listing offer

        :param listing_id: The ID of the listing to accept the offer on
        :param address_or_offerror: The address of the user to accept the offer for
        :return: The transaction receipt
        """
        self._validate_listing(listing_id)

        raw_offer = self._contract_wrapper._contract_abi.offers.call(
            listing_id, address_or_offerror
        )
        offer = ContractOffer(*raw_offer)

        return self._contract_wrapper.send_transaction(
            "accept_offer",
            [listing_id, address_or_offerror, offer.currency, offer.price_per_token],
        )

    def buyout_listing(
        self, listing_id: int, quantity_desired: int, receiver: Optional[str] = None
    ) -> TxReceipt:
        """
        Buyout a direct listing by ID

        :param listing_id: The ID of the listing to buyout
        :param quantity_desired: The quantity desired
        :param receiver: The address of the user to receive the tokens
        :return: The transaction receipt
        """
        listing = self._validate_listing(listing_id)
        valid = self._is_still_valid_listing(listing, quantity_desired)

        if not valid:
            raise Exception(
                "The asset on this listing has been moved from the lister's wallet, this listing is now invalid"
            )

        buy_for = receiver if receiver else self._contract_wrapper.get_signer_address()

        overrides: Dict[Any, Any] = {}
        set_erc20_allowance(
            self._contract_wrapper,
            listing.buyout_price * quantity_desired,
            listing.currency_contract_address,
            overrides,
        )

        # TODO: Add OVERRIDES
        return self._contract_wrapper.send_transaction(
            "buy",
            [
                listing_id,
                buy_for,
                quantity_desired,
                listing.currency_contract_address,
                listing.buyout_price * quantity_desired,
            ],
        )

    def update_listing(self, listing: DirectListing) -> TxReceipt:
        """
        Update a direct listing

        :param listing: The listing to update
        :return: The transaction receipt
        """
        return self._contract_wrapper.send_transaction(
            "update_listing",
            [
                listing.id,
                listing.quantity,
                listing.buyout_price,
                listing.buyout_price,
                listing.currency_contract_address,
                listing.start_time_in_seconds,
                listing.seconds_until_end,
            ],
        )

    def cancel_listing(self, listing_id: int) -> TxReceipt:
        """
        Cancel a direct listing

        :param listing_id: The ID of the listing to cancel
        :return: The transaction receipt
        """
        return self._contract_wrapper.send_transaction(
            "cancel_direct_listing", [listing_id]
        )

    """
    INTERNAL FUNCTIONS
    """

    def _validate_listing(self, listing_id: int) -> DirectListing:
        try:
            return self.get_listing(listing_id)
        except:
            raise ListingNotFoundException(listing_id)

    def _map_listing(self, listing: ContractListing) -> DirectListing:
        return DirectListing(
            id=listing.listing_id,
            asset_contract_address=listing.asset_contract,
            token_id=listing.token_id,
            buyout_price=listing.buyout_price_per_token,
            currency_contract_address=listing.currency,
            buyout_currency_value_per_token=fetch_currency_value(
                self._contract_wrapper.get_provider(),
                listing.currency,
                listing.buyout_price_per_token,
            ),
            quantity=listing.quantity,
            start_time_in_seconds=listing.start_time,
            asset=fetch_token_metadata_for_contract(
                listing.asset_contract,
                self._contract_wrapper.get_provider(),
                listing.token_id,
                self._storage,
            ),
            seconds_until_end=listing.end_time,
            seller_address=listing.token_owner,
        )

    def _is_still_valid_listing(
        self, listing: DirectListing, quantity: Optional[int] = None
    ) -> bool:
        approved = is_token_approved_for_marketplace(
            self._contract_wrapper.get_provider(),
            self.get_address(),
            listing.asset_contract_address,
            listing.token_id,
            listing.seller_address,
        )

        if not approved:
            return False

        provider = self._contract_wrapper.get_provider()
        erc165 = IERC165(provider, listing.asset_contract_address)
        is_erc721 = erc165.supports_interface.call(INTERFACE_ID_IERC721)
        is_erc1155 = erc165.supports_interface.call(INTERFACE_ID_IERC1155)

        if is_erc721:
            ierc721 = IERC721(provider, listing.asset_contract_address)
            return (
                ierc721.owner_of.call(listing.token_id).lower()
                == listing.seller_address.lower()
            )
        elif is_erc1155:
            ierc1155 = IERC1155(provider, listing.asset_contract_address)
            balance = ierc1155.balance_of.call(listing.seller_address, listing.token_id)
            return balance > (quantity if quantity is not None else listing.quantity)
        else:
            return False



from dataclasses import dataclass
from typing import List


@dataclass
class CidWithFileName:
    cid: str
    filenames: List[str]


@dataclass
class UriWithMetadata:
    base_uri: str
    metadata_uris: List[str]



from typing import Any


class NotFoundException(Exception):
    def __init__(self, indentifier: str):
        super().__init__(f"NOT FOUND: Object with ID {indentifier} not found")


class NoSignerException(Exception):
    def __init__(self):
        super().__init__(
            (
                "You need to pass a valid signer to the SDK to use this function. "
                "Please use the `update_signer` method of the SDK to set a valid signer. "
                "You can create a signer using the `Account.from_key(<private_key>)` method."
            )
        )


class UploadException(Exception):
    def __init__(self, message: str):
        super().__init__(f"UPLOAD FAILED: {message}")


class NotEnoughTokensException(Exception):
    def __init__(self, contract_address: str, quantity: str, available: int):
        super().__init__(
            f"BALANCE ERROR: you do not have enough balance on contract ${contract_address} to use ${quantity} tokens. You have ${available} tokens available."
        )


class FetchException(Exception):
    def __init__(self, message: str):
        super().__init__(f"FETCH FAILED: {message}")


class RestrictedTransferException(Exception):
    def __init__(self, asset_address: str):
        super().__init__(
            f"Failed to transfer asset, transfer is restricted for asset {asset_address}"
        )


class DuplicateFileNameException(Exception):
    def __init__(self, filename: str):
        super().__init__(f"File with name {filename} already exists")


class RoleException(Exception):
    def __init__(self, message: str):
        super().__init__(f"ROLE ERROR: {message}")


class WrongListingTypeException(Exception):
    def __init__(self, listing_id: int, listing_type: str, expected_type: str):
        super().__init__(
            f"The listing type of {str(listing_id)} is not valid : {listing_type}, expected {expected_type}."
        )


class AuctionAlreadyStartedException(Exception):
    def __init__(self, listing_id: int):
        super().__init__(
            f"The auction of listing {str(listing_id)} has already started."
        )


class AuctionHasNotEndedException(Exception):
    def __init__(self, listing_id: int):
        super().__init__(f"The auction of listing {str(listing_id)} has not ended yet.")


class ListingNotFoundException(Exception):
    def __init__(self, listing_id: int):
        super().__init__(f"Error getting the listing with ID {str(listing_id)}.")


class DuplicateLeafsException(Exception):
    def __init__(self, message: str = ""):
        super().__init__(f"Duplicate leafs: {message}")


def includes_error_message(err: Any, message: str) -> bool:
    return message in str(err)



from time import time
from typing import Any, Dict, Optional, cast
from thirdweb.abi import Marketplace
from thirdweb.abi.marketplace import IMarketplaceListingParameters
from thirdweb.common.currency import (
    fetch_currency_metadata,
    fetch_currency_value,
    format_units,
    normalize_price_value,
    set_erc20_allowance,
)
from thirdweb.common.error import (
    AuctionAlreadyStartedException,
    ListingNotFoundException,
    WrongListingTypeException,
)
from thirdweb.common.marketplace import (
    handle_token_approval,
    is_winning_bid,
    map_offer,
    validate_new_listing_param,
)
from thirdweb.common.nft import fetch_token_metadata_for_contract
from thirdweb.constants.currency import ZERO_ADDRESS
from thirdweb.core.classes.base_contract import BaseContract
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.types.currency import Price
from thirdweb.types.marketplace import (
    AuctionListing,
    ContractListing,
    ContractOffer,
    ListingType,
    NewAuctionListing,
    Offer,
)
from web3.eth import TxReceipt


class MarketplaceAuction(BaseContract[Marketplace]):
    _storage: IpfsStorage

    def __init__(self, contract_wrapper: ContractWrapper, storage: IpfsStorage):
        super().__init__(contract_wrapper)
        self._storage = storage

    """
    READ FUNCTIONS
    """

    def get_listing(self, listing_id: int) -> AuctionListing:
        """
        Get an auction listing from the marketplace by ID

        :param listing_id: The ID of the listing to get
        :return: The auction listing
        """
        raw_listing = self._contract_wrapper._contract_abi.listings.call(listing_id)
        listing = ContractListing(*raw_listing)

        if listing.listing_id != listing_id:
            raise ListingNotFoundException(listing_id)

        if ListingType(listing.listing_type) != ListingType.AUCTION:
            raise WrongListingTypeException(
                listing_id,
                "Auction",
                "Auction",
            )

        return self._map_listing(listing)

    def get_winning_bid(self, listing_id: int) -> Optional[Offer]:
        """
        Get the winning bid on an auction listing

        :param listing_id: The ID of the listing to get the winning bid for
        :return: The winning bid
        """
        self._validate_listing(listing_id)
        raw_offers = self._contract_wrapper._contract_abi.winning_bid.call(listing_id)
        offers = ContractOffer(*raw_offers)

        if offers.offeror == ZERO_ADDRESS:
            return None

        return map_offer(self._contract_wrapper.get_provider(), listing_id, offers)

    def get_winner(self, listing_id) -> str:
        """
        Get the winner of an auction that has already ended.

        :param listing_id: The ID of the listing to get the winner for
        :return: The winning bidder
        """
        interface = self._contract_wrapper.get_contract_interface()
        closed_auctions = interface.events.AuctionClosed().getLogs()

        auctions = [
            auction.get("args").get("listing_id") == listing_id
            for auction in closed_auctions
        ]

        if len(auctions) == 0:
            raise Exception(
                f"Could not find auction with listing ID {listing_id} in close auctions"
            )

        return auctions[0].get("winning_bidder")

    """
    WRITE FUNCTIONS
    """

    def create_listing(self, listing: NewAuctionListing) -> int:
        """
        Create a new auction listing

        :param listing: The listing to create
        :return: The ID of the listing
        """
        validate_new_listing_param(listing)

        handle_token_approval(
            self._contract_wrapper.get_provider(),
            self._contract_wrapper.get_signer(),
            self.get_address(),
            listing.asset_contract_address,
            listing.token_id,
            self._contract_wrapper.get_signer_address(),
        )

        normalized_price_per_token = normalize_price_value(
            self._contract_wrapper.get_provider(),
            listing.buyout_price_per_token,
            listing.currency_contract_address,
        )

        normalized_reserve_price = normalize_price_value(
            self._contract_wrapper.get_provider(),
            listing.reserve_price_per_token,
            listing.currency_contract_address,
        )

        receipt = self._contract_wrapper.send_transaction(
            "create_listing",
            [
                IMarketplaceListingParameters(
                    assetContract=listing.asset_contract_address,
                    tokenId=listing.token_id,
                    startTime=listing.start_time_in_seconds,
                    secondsUntilEndTime=listing.listing_duration_in_seconds,
                    quantityToList=listing.quantity,
                    currencyToAccept=listing.currency_contract_address,
                    reservePricePerToken=normalized_reserve_price,
                    buyoutPricePerToken=normalized_price_per_token,
                    listingType=1,
                ),
            ],
        )

        events = self._contract_wrapper.get_events("ListingAdded", receipt)
        return cast(Any, events[0].get("args")).get("listingId")

    def buyout_listing(self, listing_id: int) -> TxReceipt:
        """
        Buyout an auction listing

        :param listing_id: The ID of the listing to buyout
        :return: The transaction receipt
        """
        listing = self._validate_listing(listing_id)

        currency_metadata = fetch_currency_metadata(
            self._contract_wrapper.get_provider(), listing.currency_contract_address
        )

        return self.make_bid(
            listing_id, format_units(listing.buyout_price, currency_metadata.decimals)
        )

    def make_bid(self, listing_id: int, price_per_token: Price) -> TxReceipt:
        """
        Make a bid on an auction listing

        :param listing_id: The ID of the listing to bid on
        :param price_per_token: The price per token to bid
        :return: The transaction receipt
        """
        listing = self._validate_listing(listing_id)
        normalized_price = normalize_price_value(
            self._contract_wrapper.get_provider(),
            price_per_token,
            listing.currency_contract_address,
        )

        bid_buffer = self._contract_wrapper._contract_abi.bid_buffer_bps.call()
        winning_bid = self.get_winning_bid(listing_id)

        if winning_bid is not None:
            is_winner = is_winning_bid(
                winning_bid.price_per_token, normalized_price, bid_buffer
            )

            if not is_winner:
                "Bid price is too low based on the current winning bid and the bid buffer"
        else:
            token_price = normalized_price
            reserve_price = listing.reserve_price

            if not token_price >= reserve_price:
                "Bid price is too low based on reserve price"

        quantity = listing.quantity
        value = normalized_price * quantity

        overrides: Dict[Any, Any] = {}
        set_erc20_allowance(
            self._contract_wrapper, value, listing.currency_contract_address, overrides
        )

        return self._contract_wrapper.send_transaction(
            "offer",
            [
                listing_id,
                listing.quantity,
                listing.currency_contract_address,
                normalized_price,
            ],
        )

    def cancel_listing(self, listing_id: int) -> TxReceipt:
        """
        Cancel an auction listing

        :param listing_id: The ID of the listing to cancel
        :return: The transaction receipt
        """
        listing = self._validate_listing(listing_id)

        now = int(time())
        start_time = listing.start_time_in_epoch_seconds

        raw_offers = self._contract_wrapper._contract_abi.winning_bid.call(listing_id)
        offers = ContractOffer(*raw_offers)

        if now > start_time and offers.offeror != ZERO_ADDRESS:
            raise AuctionAlreadyStartedException(listing_id)

        return self._contract_wrapper.send_transaction(
            "close_auction", [listing_id, self._contract_wrapper.get_signer_address()]
        )

    def close_listing(
        self, listing_id: int, close_for: Optional[str] = None
    ) -> TxReceipt:
        if not close_for:
            close_for = self._contract_wrapper.get_signer_address()

        listing = self._validate_listing(listing_id)

        try:
            return self._contract_wrapper.send_transaction(
                "close_auction", [listing_id, close_for]
            )
        except Exception as err:
            raise Exception(err)

    def update_listing(self, listing: AuctionListing) -> TxReceipt:
        """
        Update a listing on the marketplace

        :param listing: The listing to update
        :return: The transaction receipt
        """

        return self._contract_wrapper.send_transaction(
            "update_listing",
            [
                listing.id,
                listing.quantity,
                listing.reserve_price,
                listing.buyout_price,
                listing.currency_contract_address,
                listing.start_time_in_epoch_seconds,
                listing.end_time_in_epoch_seconds,
            ],
        )

    """
    INTERNAL FUNCTIONS
    """

    def _validate_listing(self, listing_id: int) -> AuctionListing:
        try:
            return self.get_listing(listing_id)
        except:
            raise ListingNotFoundException(listing_id)

    def _map_listing(self, listing: ContractListing) -> AuctionListing:
        return AuctionListing(
            id=listing.listing_id,
            asset_contract_address=listing.asset_contract,
            buyout_price=listing.buyout_price_per_token,
            currency_contract_address=listing.currency,
            buyout_currency_value_per_token=fetch_currency_value(
                self._contract_wrapper.get_provider(),
                listing.currency,
                listing.buyout_price_per_token,
            ),
            token_id=listing.token_id,
            quantity=listing.quantity,
            start_time_in_epoch_seconds=listing.start_time,
            asset=fetch_token_metadata_for_contract(
                listing.asset_contract,
                self._contract_wrapper.get_provider(),
                listing.token_id,
                self._storage,
            ),
            reserve_price_currency_value_per_token=fetch_currency_value(
                self._contract_wrapper.get_provider(),
                listing.currency,
                listing.reserve_price_per_token,
            ),
            reserve_price=listing.reserve_price_per_token,
            end_time_in_epoch_seconds=listing.end_time,
            seller_address=listing.token_owner,
        )



from dataclasses import dataclass
from enum import Enum
from typing import Optional
from thirdweb.types.currency import CurrencyValue, Price, PriceWei

from thirdweb.types.nft import NFTMetadata


class ListingType(Enum):
    DIRECT = 0
    AUCTION = 1


@dataclass
class MarketplaceFilter:
    start: int = 0
    count: int = 100
    seller: Optional[str] = None
    token_contract: Optional[str] = None
    token_id: Optional[int] = None


@dataclass
class NewDirectListing:
    type = "NewDirectListing"
    asset_contract_address: str
    token_id: int
    start_time_in_seconds: int
    listing_duration_in_seconds: int
    quantity: int
    currency_contract_address: str
    buyout_price_per_token: int


@dataclass
class NewAuctionListing:
    type = "NewAuctionListing"
    asset_contract_address: str
    token_id: int
    start_time_in_seconds: int
    listing_duration_in_seconds: int
    quantity: int
    currency_contract_address: str
    reserve_price_per_token: int
    buyout_price_per_token: int


@dataclass
class ContractListing:
    listing_id: int
    token_owner: str
    asset_contract: str
    token_id: int
    start_time: int
    end_time: int
    quantity: int
    currency: str
    reserve_price_per_token: PriceWei
    buyout_price_per_token: PriceWei
    token_type: int
    listing_type: int


@dataclass
class DirectListing:
    id: int
    asset_contract_address: str
    token_id: int
    asset: NFTMetadata
    start_time_in_seconds: int
    seconds_until_end: int
    quantity: int
    currency_contract_address: str
    buyout_currency_value_per_token: CurrencyValue
    buyout_price: PriceWei
    seller_address: str
    type = ListingType.DIRECT


@dataclass
class AuctionListing:
    id: int
    asset_contract_address: str
    token_id: int
    asset: NFTMetadata
    start_time_in_epoch_seconds: int
    end_time_in_epoch_seconds: int
    quantity: int
    currency_contract_address: str
    reserve_price: PriceWei
    buyout_price: PriceWei
    buyout_currency_value_per_token: CurrencyValue
    reserve_price_currency_value_per_token: CurrencyValue
    seller_address: str
    type = ListingType.AUCTION


@dataclass
class ContractOffer:
    listing_id: int
    offeror: str
    quantity_wanted: int
    currency: str
    price_per_token: PriceWei
    expiration_timestamp: int


@dataclass
class Offer:
    listing_id: int
    buyer_address: str
    quantity_desired: int
    price_per_token: PriceWei
    currency_value: CurrencyValue
    currency_contract_address: str



from enum import Enum
from typing import Optional
from dataclasses import dataclass

from thirdweb.constants.chains import ChainId


class GasSpeed(Enum):
    STANDARD = "standard"
    FAST = "fast"
    FASTEST = "fastest"


@dataclass
class GasSettings(object):
    """
    The gas settings for the SDK.

    :param max_price_in_gwei: maximum gas price in gwei, defaults to 300
    :param speed: gas speed to use, defaults to "fastest"
    """

    max_price_in_gwei: int = 300
    speed: GasSpeed = GasSpeed.FASTEST


@dataclass
class ReadOnlySettings(object):
    """
    The read-only RPC settings for the SDK.

    :param rpc_url: URL of the RPC
    :param chain_id: optional chain ID to use for the RPC
    """

    rpc_url: str = ""
    chain_id: Optional[ChainId] = None


@dataclass
class SDKOptions(object):
    """
    Optional settings to configure the SDK

    :param read_only_settings: optional read-only RPC settings
    :param gas_settings: gas settings
    """

    read_only_settings: Optional[ReadOnlySettings] = None
    gas_settings: GasSettings = GasSettings()



from typing import List, Optional, cast
from eth_utils import is_address
from web3 import Web3
from eth_account.account import LocalAccount
from thirdweb.constants.currency import ZERO_ADDRESS
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.abi import TWRegistry
from thirdweb.types.sdk import SDKOptions


class ContractRegistry(ContractWrapper):
    def __init__(
        self,
        registry_address: str,
        provider: Web3,
        signer: Optional[LocalAccount],
        options: SDKOptions = SDKOptions(),
    ):
        abi = TWRegistry(provider, registry_address)
        super().__init__(abi, provider, signer, options)

    def get_contract_addresses(self, address: str) -> List[str]:
        """
        Get all the contract addresses registered for a given address.

        :param address: address to get the contract addresses for
        :returns: list of contract addresses
        """

        contract_addresses = self._get_abi().get_all.call(address)
        return [
            address
            for address in contract_addresses
            if address != ZERO_ADDRESS and is_address(address)
        ]

    def _get_abi(self) -> TWRegistry:
        return cast(TWRegistry, self._contract_abi)



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

from web3 import Web3
from thirdweb.abi.drop_erc721 import IDropClaimConditionClaimCondition
from thirdweb.common.snapshots import create_snapshot
from thirdweb.constants.addresses import DEFAULT_MERKLE_ROOT
from thirdweb.constants.currency import NATIVE_TOKEN_ADDRESS, ZERO_ADDRESS
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.common.currency import (
    approve_erc20_allowance,
    fetch_currency_value,
    is_native_token,
    normalize_price_value,
    parse_units,
)

from thirdweb.types.contracts.claim_conditions import (
    ClaimCondition,
    ClaimConditionInput,
    ClaimConditionOutput,
    ClaimVerification,
    FilledConditionInput,
    Snapshot,
    SnapshotInfo,
    SnapshotProof,
)


def prepare_claim(
    quantity: int,
    active_claim_condition: ClaimCondition,
    merkle_metadata: Dict[str, str],
    contract_wrapper: ContractWrapper,
    storage: IpfsStorage,
    proofs: List[str] = [DEFAULT_MERKLE_ROOT],
) -> ClaimVerification:
    address_to_claim = contract_wrapper.get_signer_address()
    token_decimals = 0

    max_claimable = 0

    try:
        if not active_claim_condition.merkle_root_hash.startswith(ZERO_ADDRESS):
            claims = fetch_snapshot(
                active_claim_condition.merkle_root_hash, merkle_metadata, storage
            )
            item = (
                next((c for c in claims if c.address == address_to_claim), None)
                if claims is not None
                else None
            )
            if item is None:
                raise Exception("No claim found for this address")
            proofs = item.proof
            max_claimable = parse_units(item.max_claimable, token_decimals)
    except Exception as e:
        if str(e) == "No claim found for this address":
            raise e

        print(
            "Failed to check claim condition merkle root hash, continuing anyway: ",
            str(e),
        )

    # TODO: OVERRIDES
    overrides: Dict[str, Any] = {}
    price = active_claim_condition.price
    currency_address = active_claim_condition.currency_address

    if price > 0:
        if is_native_token(currency_address):
            # overrides["value"] = quantity * price
            pass
        else:
            approve_erc20_allowance(contract_wrapper, currency_address, price, quantity)

    return ClaimVerification(
        proofs=proofs,
        max_quantity_per_transaction=max_claimable,
        price=price,
        currency_address=currency_address,
    )


def fetch_snapshot(
    merkle_root: str,
    merkle_metadata: Dict[str, str],
    storage: IpfsStorage,
) -> Optional[List[SnapshotProof]]:
    snapshot = None
    if not merkle_root in merkle_metadata:
        return snapshot

    snapshot_uri = merkle_metadata[merkle_root]
    if snapshot_uri:
        raw = storage.get(snapshot_uri)
        snapshot_data: Snapshot = Snapshot.from_json(raw)
        if merkle_root.lower() == snapshot_data.merkle_root.lower():
            snapshot = snapshot_data.claims

    return snapshot


@dataclass
class ClaimerProof:
    max_claimable: int
    proof: List[str]


def get_claimer_proofs(
    address_to_claim: str,
    merkle_root: str,
    token_decimals: int,
    merkle_metadata: Dict[str, str],
    storage: IpfsStorage,
) -> ClaimerProof:
    claims = fetch_snapshot(merkle_root, merkle_metadata, storage)

    if claims is None:
        return ClaimerProof(max_claimable=0, proof=[])
    item = next((c for c in claims if c.address == address_to_claim), None)

    if item is None:
        return ClaimerProof(proof=[], max_claimable=0)

    return ClaimerProof(
        proof=item.proof, max_claimable=parse_units(item.max_claimable, token_decimals)
    )


def process_claim_condition_inputs(
    claim_condition_inputs: List[ClaimConditionInput],
    token_decimals: int,
    provider: Web3,
    storage: IpfsStorage,
) -> Tuple[List[SnapshotInfo], List[IDropClaimConditionClaimCondition]]:
    snapshot_infos: List[SnapshotInfo] = []
    inputs_with_snapshots: List[ClaimConditionInput] = []
    for condition_input in claim_condition_inputs:
        if condition_input.snapshot:
            snapshot_info = create_snapshot(
                condition_input.snapshot, token_decimals, storage
            )
            snapshot_infos.append(snapshot_info)
            condition_input.merkle_root_hash = snapshot_info.merkle_root

        inputs_with_snapshots.append(condition_input)

    parsed_inputs = inputs_with_snapshots

    sorted_conditions = sorted(
        [convert_to_contract_model(c, token_decimals, provider) for c in parsed_inputs],
        key=lambda x: x["startTimestamp"],
    )

    return (snapshot_infos, sorted_conditions)


def convert_to_contract_model(
    c: FilledConditionInput, token_decimals: int, provider: Web3
) -> IDropClaimConditionClaimCondition:
    currency = (
        NATIVE_TOKEN_ADDRESS
        if ZERO_ADDRESS == c.currency_address
        else c.currency_address
    )

    # TODO: Include unlimited here and change types
    max_claimable_supply = parse_units(c.max_quantity, token_decimals)
    quantity_limit_per_transaction = parse_units(
        c.quantity_limit_per_transaction, token_decimals
    )

    return IDropClaimConditionClaimCondition(
        startTimestamp=c.start_time,
        maxClaimableSupply=max_claimable_supply,
        supplyClaimed=0,
        quantityLimitPerTransaction=quantity_limit_per_transaction,
        waitTimeInSecondsBetweenClaims=c.wait_in_seconds,
        pricePerToken=normalize_price_value(provider, c.price, currency),
        currency=currency,
        merkleRoot=c.merkle_root_hash,
    )


def transform_result_to_claim_condition(
    pm: IDropClaimConditionClaimCondition,
    provider: Web3,
    merkle_metadata: Dict[str, str],
    storage: IpfsStorage,
) -> ClaimCondition:
    cv = fetch_currency_value(provider, pm["currency"], pm["pricePerToken"])
    claims = fetch_snapshot(str(pm["merkleRoot"]), merkle_metadata, storage)
    return ClaimConditionOutput(
        start_time=pm["startTimestamp"],
        max_quantity=pm["maxClaimableSupply"],
        available_supply=pm["maxClaimableSupply"] - pm["supplyClaimed"],
        quantity_limit_per_tranaction=pm["quantityLimitPerTransaction"],
        wait_in_seconds=pm["waitTimeInSecondsBetweenClaims"],
        price=pm["pricePerToken"],
        currency_address=pm["currency"],
        currency_metadata=cv,
        merkle_root_hash=pm["merkleRoot"],
        snapshot=claims,
    )



from typing import Optional
from uuid import uuid4


def resolve_or_generate_id(request_uid: Optional[str]) -> bytes:
    # TODO: Implement encoding
    if request_uid is None:
        generated_id = uuid4().hex
        return str.encode(generated_id)
    return str.encode(request_uid)



from typing import cast

from eth_typing import Address
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from web3.contract import ContractFunctions
from web3 import Web3
from thirdweb.core.classes.ipfs_storage import IpfsStorage

from thirdweb.types.contract import TContractABI
from thirdweb.types.settings.metadata import CustomContractMetadata


def matches_interface(
    contract: ContractFunctions, interface_to_match: ContractFunctions
) -> bool:
    contract_functions = [fn for fn in contract]
    interface_fn = [fn for fn in interface_to_match]

    overlap = set(contract_functions).intersection(interface_fn)
    return set(interface_fn) == set(overlap)


def fetch_contract_metadata(metadata_uri: str, storage: IpfsStorage) -> str:
    metadata = storage.get(metadata_uri)
    abi = storage._get(metadata["abiUri"]).json()
    return abi  # type: ignore



"""Generated wrapper for IERC165 Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for IERC165 below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        IERC165Validator,
    )
except ImportError:

    class IERC165Validator(  # type: ignore
        Validator
    ):
        """No-op input validator."""


try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass





class SupportsInterfaceMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the supportsInterface method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, interface_id: Union[bytes, str]):
        """Validate the inputs to the supportsInterface method."""
        self.validator.assert_valid(
            method_name='supportsInterface',
            parameter_name='interfaceId',
            argument_value=interface_id,
        )
        return (interface_id)

    def call(self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(interface_id).call(tx_params.as_dict())
        return bool(returned)

    def send_transaction(self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(interface_id).transact(tx_params.as_dict())

    def build_transaction(self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(interface_id).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, interface_id: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (interface_id) = self.validate_and_normalize_inputs(interface_id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(interface_id).estimateGas(tx_params.as_dict())

# pylint: disable=too-many-public-methods,too-many-instance-attributes
class IERC165:
    """Wrapper class for IERC165 Solidity contract."""
    supports_interface: SupportsInterfaceMethod
    """Constructor-initialized instance of
    :class:`SupportsInterfaceMethod`.
    """


    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        validator: IERC165Validator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param web3_or_provider: Either an instance of `web3.Web3`:code: or
            `web3.providers.base.BaseProvider`:code:
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        # pylint: disable=too-many-statements

        self.contract_address = contract_address

        if not validator:
            validator = IERC165Validator(web3_or_provider, contract_address)

        web3 = None
        if isinstance(web3_or_provider, BaseProvider):
            web3 = Web3(web3_or_provider)
        elif isinstance(web3_or_provider, Web3):
            web3 = web3_or_provider
        else:
            raise TypeError(
                "Expected parameter 'web3_or_provider' to be an instance of either"
                + " Web3 or BaseProvider"
            )

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                         middleware['function'], layer=middleware['layer'],
                    )
            except ValueError as value_error:
                if value_error.args == ("You can't add the same un-named instance twice",):
                    pass

        self._web3_eth = web3.eth

        functions = self._web3_eth.contract(address=to_checksum_address(contract_address), abi=IERC165.abi()).functions

        self.supports_interface = SupportsInterfaceMethod(web3_or_provider, contract_address, functions.supportsInterface, validator)


    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"inputs":[{"internalType":"bytes4","name":"interfaceId","type":"bytes4"}],"name":"supportsInterface","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"}]'  # noqa: E501 (line-too-long)
        )

# pylint: disable=too-many-lines



from typing import Dict, List, Optional, Any, cast
from eth_typing import Address
from eth_utils import encode_hex
from web3 import Web3
from thirdweb.abi.t_w_factory import TWFactory
from thirdweb.constants.addresses import (
    CONTRACT_ADDRESSES,
    OZ_DEFENDER_FORWARDER_ADDRESS,
)
from thirdweb.constants.chains import ChainId
from thirdweb.constants.currency import ZERO_ADDRESS
from thirdweb.contracts import (
    NFTCollection,
    Edition,
    Token,
    NFTDrop,
    EditionDrop,
    Marketplace,
)
from thirdweb.contracts.maps import (
    CONTRACTS_MAP,
    REMOTE_CONTRACT_NAME,
)
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from eth_account.account import LocalAccount

from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.types.contract import ContractType
from thirdweb.types.sdk import SDKOptions
from thirdweb.types.settings.metadata import (
    NFTDropContractMetadata,
    EditionDropContractMetadata,
    MarketplaceContractMetadata,
    TokenContractMetadata,
)


class ContractFactory(ContractWrapper[TWFactory]):
    _storage: IpfsStorage

    def __init__(
        self,
        factory_address: str,
        provider: Web3,
        signer: Optional[LocalAccount] = None,
        options: SDKOptions = SDKOptions(),
        storage: IpfsStorage = IpfsStorage(),
    ):
        abi = TWFactory(provider, factory_address)
        super().__init__(abi, provider, signer, options)
        self._storage = storage

    def deploy(
        self, contract_type: ContractType, contract_metadata: Dict[str, Any]
    ) -> str:
        # First, we upload the contract metadata to IPFS
        contract_uri = self._storage.upload_metadata(
            contract_metadata,
            self._contract_abi.contract_address,
            self.get_signer_address(),
        )

        # Then, we get setup the contract interface for the contract we want to deploy
        contract = CONTRACTS_MAP[contract_type]
        interface = self.get_provider().eth.contract(
            address=cast(Address, self._contract_abi.contract_address),
            abi=contract._abi_type.abi(),
        )

        # Next, we encode all the initialize function arguments and pass them to the proxy deploy
        deploy_arguments = self.get_deploy_arguments(
            contract_type, contract_metadata, contract_uri
        )

        encoded_function = interface.encodeABI(
            fn_name="initialize", args=deploy_arguments
        )
        contract_name = REMOTE_CONTRACT_NAME[contract_type]
        encoded_type = encode_hex(contract_name).ljust(66, "0")
        receipt = self.send_transaction(
            "deploy_proxy", [encoded_type, encoded_function]
        )

        events = self.get_events("ProxyDeployed", receipt)
        if len(events) == 0 or events[0].get("event") != "ProxyDeployed":
            raise Exception("No proxy deployed event found")

        address = cast(Any, events[0].get("args")).get("proxy")
        return address

    def get_deploy_arguments(
        self,
        contract_type: ContractType,
        contract_metadata: Dict[str, Any],
        contract_uri: str,
    ) -> List[Any]:
        trusted_forwarders = self._get_default_trusted_forwarders()
        if "trusted_forwarders" in contract_metadata:
            trusted_forwarders = contract_metadata["trusted_forwarders"]

        if (
            contract_type == NFTCollection.contract_type
            or contract_type == NFTDrop.contract_type
        ):
            metadata = NFTDropContractMetadata.from_json(contract_metadata)
            return [
                self.get_signer_address(),
                metadata.name,
                metadata.symbol,
                contract_uri,
                trusted_forwarders,
                metadata.primary_sale_recipient,
                metadata.fee_recipient,
                metadata.seller_fee_basis_points,
                metadata.platform_fee_basis_points,
                metadata.platform_fee_recipient,
            ]

        if (
            contract_type == Edition.contract_type
            or contract_type == EditionDrop.contract_type
        ):
            metadata = EditionDropContractMetadata.from_json(contract_metadata)
            return [
                self.get_signer_address(),
                metadata.name,
                metadata.symbol,
                contract_uri,
                trusted_forwarders,
                metadata.primary_sale_recipient,
                metadata.fee_recipient,
                metadata.seller_fee_basis_points,
                metadata.platform_fee_basis_points,
                metadata.platform_fee_recipient,
            ]

        if contract_type == Token.contract_type:
            metadata = TokenContractMetadata.from_json(contract_metadata)
            return [
                self.get_signer_address(),
                metadata.name,
                metadata.symbol,
                contract_uri,
                trusted_forwarders,
                metadata.primary_sale_recipient,
                metadata.platform_fee_recipient,
                metadata.platform_fee_basis_points,
            ]
        if contract_type == Marketplace.contract_type:
            metadata = MarketplaceContractMetadata.from_json(contract_metadata)

            return [
                self.get_signer_address(),
                contract_uri,
                trusted_forwarders,
                metadata.platform_fee_recipient,
                metadata.platform_fee_basis_points,
            ]

        return []

    def _get_default_trusted_forwarders(self) -> List[str]:
        chain_id = ChainId(self.get_chain_id())

        if chain_id in CONTRACT_ADDRESSES:
            biconomy_forwarder_address = CONTRACT_ADDRESSES[chain_id][
                "biconomy_forwarder"
            ]

            if biconomy_forwarder_address != ZERO_ADDRESS:
                return [OZ_DEFENDER_FORWARDER_ADDRESS, biconomy_forwarder_address]
        return [OZ_DEFENDER_FORWARDER_ADDRESS]



from dataclasses import dataclass
from typing import Any, Dict, List, Union
from eth_account.account import LocalAccount
from thirdweb_eth_account.messages import encode_structured_data
from web3 import Web3

from thirdweb.types.contracts.signature import SignedPayload721


@dataclass
class EIP712StandardDomain:
    name: str
    version: str
    chainId: int
    verifyingContract: str

    def to_json(self) -> Dict[str, Any]:
        return self.__dict__


@dataclass
class EIP712PolygonDomain:
    name: str
    version: str
    verifyingContract: str
    salt: str

    def to_json(self) -> Dict[str, Any]:
        return self.__dict__


EIP712Domain = Union[EIP712StandardDomain, EIP712PolygonDomain]


def sign_typed_data_internal(
    provider: Web3,
    signer: LocalAccount,
    domain: EIP712Domain,
    types: Dict[str, List[Any]],
    message: Dict[str, Any],
) -> bytes:
    encoded_message = encode_structured_data(
        {
            "domain": domain.to_json(),
            "types": types,
            "primaryType": "MintRequest",
            "message": message,
        }
    )

    signature = provider.eth.account.sign_message(
        encoded_message, signer._private_key
    ).signature.hex()
    return signature



"""Generated wrapper for IPermissionsEnumerable Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for IPermissionsEnumerable below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        IPermissionsEnumerableValidator,
    )
except ImportError:

    class IPermissionsEnumerableValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""


try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass





class GetRoleAdminMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the getRoleAdmin method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str]):
        """Validate the inputs to the getRoleAdmin method."""
        self.validator.assert_valid(
            method_name='getRoleAdmin',
            parameter_name='role',
            argument_value=role,
        )
        return (role)

    def call(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[bytes, str]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role).call(tx_params.as_dict())
        return Union[bytes, str](returned)

    def send_transaction(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).estimateGas(tx_params.as_dict())

class GetRoleMemberMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the getRoleMember method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], index: int):
        """Validate the inputs to the getRoleMember method."""
        self.validator.assert_valid(
            method_name='getRoleMember',
            parameter_name='role',
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name='getRoleMember',
            parameter_name='index',
            argument_value=index,
        )
        # safeguard against fractional inputs
        index = int(index)
        return (role, index)

    def call(self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role, index).call(tx_params.as_dict())
        return str(returned)

    def send_transaction(self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, index).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, index).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], index: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role, index) = self.validate_and_normalize_inputs(role, index)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, index).estimateGas(tx_params.as_dict())

class GetRoleMemberCountMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the getRoleMemberCount method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str]):
        """Validate the inputs to the getRoleMemberCount method."""
        self.validator.assert_valid(
            method_name='getRoleMemberCount',
            parameter_name='role',
            argument_value=role,
        )
        return (role)

    def call(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role).call(tx_params.as_dict())
        return int(returned)

    def send_transaction(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role) = self.validate_and_normalize_inputs(role)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role).estimateGas(tx_params.as_dict())

class GrantRoleMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the grantRole method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the grantRole method."""
        self.validator.assert_valid(
            method_name='grantRole',
            parameter_name='role',
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name='grantRole',
            parameter_name='account',
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(role, account).call(tx_params.as_dict())

    def send_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())

class HasRoleMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the hasRole method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the hasRole method."""
        self.validator.assert_valid(
            method_name='hasRole',
            parameter_name='role',
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name='hasRole',
            parameter_name='account',
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(role, account).call(tx_params.as_dict())
        return bool(returned)

    def send_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())

class RenounceRoleMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the renounceRole method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the renounceRole method."""
        self.validator.assert_valid(
            method_name='renounceRole',
            parameter_name='role',
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name='renounceRole',
            parameter_name='account',
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(role, account).call(tx_params.as_dict())

    def send_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())

class RevokeRoleMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the revokeRole method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, role: Union[bytes, str], account: str):
        """Validate the inputs to the revokeRole method."""
        self.validator.assert_valid(
            method_name='revokeRole',
            parameter_name='role',
            argument_value=role,
        )
        self.validator.assert_valid(
            method_name='revokeRole',
            parameter_name='account',
            argument_value=account,
        )
        account = self.validate_and_checksum_address(account)
        return (role, account)

    def call(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(role, account).call(tx_params.as_dict())

    def send_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).transact(tx_params.as_dict())

    def build_transaction(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, role: Union[bytes, str], account: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (role, account) = self.validate_and_normalize_inputs(role, account)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(role, account).estimateGas(tx_params.as_dict())

# pylint: disable=too-many-public-methods,too-many-instance-attributes
class IPermissionsEnumerable:
    """Wrapper class for IPermissionsEnumerable Solidity contract."""
    get_role_admin: GetRoleAdminMethod
    """Constructor-initialized instance of
    :class:`GetRoleAdminMethod`.
    """

    get_role_member: GetRoleMemberMethod
    """Constructor-initialized instance of
    :class:`GetRoleMemberMethod`.
    """

    get_role_member_count: GetRoleMemberCountMethod
    """Constructor-initialized instance of
    :class:`GetRoleMemberCountMethod`.
    """

    grant_role: GrantRoleMethod
    """Constructor-initialized instance of
    :class:`GrantRoleMethod`.
    """

    has_role: HasRoleMethod
    """Constructor-initialized instance of
    :class:`HasRoleMethod`.
    """

    renounce_role: RenounceRoleMethod
    """Constructor-initialized instance of
    :class:`RenounceRoleMethod`.
    """

    revoke_role: RevokeRoleMethod
    """Constructor-initialized instance of
    :class:`RevokeRoleMethod`.
    """


    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        validator: IPermissionsEnumerableValidator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param web3_or_provider: Either an instance of `web3.Web3`:code: or
            `web3.providers.base.BaseProvider`:code:
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        # pylint: disable=too-many-statements

        self.contract_address = contract_address

        if not validator:
            validator = IPermissionsEnumerableValidator(web3_or_provider, contract_address)

        web3 = None
        if isinstance(web3_or_provider, BaseProvider):
            web3 = Web3(web3_or_provider)
        elif isinstance(web3_or_provider, Web3):
            web3 = web3_or_provider
        else:
            raise TypeError(
                "Expected parameter 'web3_or_provider' to be an instance of either"
                + " Web3 or BaseProvider"
            )

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                         middleware['function'], layer=middleware['layer'],
                    )
            except ValueError as value_error:
                if value_error.args == ("You can't add the same un-named instance twice",):
                    pass

        self._web3_eth = web3.eth

        functions = self._web3_eth.contract(address=to_checksum_address(contract_address), abi=IPermissionsEnumerable.abi()).functions

        self.get_role_admin = GetRoleAdminMethod(web3_or_provider, contract_address, functions.getRoleAdmin, validator)

        self.get_role_member = GetRoleMemberMethod(web3_or_provider, contract_address, functions.getRoleMember, validator)

        self.get_role_member_count = GetRoleMemberCountMethod(web3_or_provider, contract_address, functions.getRoleMemberCount, validator)

        self.grant_role = GrantRoleMethod(web3_or_provider, contract_address, functions.grantRole, validator)

        self.has_role = HasRoleMethod(web3_or_provider, contract_address, functions.hasRole, validator)

        self.renounce_role = RenounceRoleMethod(web3_or_provider, contract_address, functions.renounceRole, validator)

        self.revoke_role = RevokeRoleMethod(web3_or_provider, contract_address, functions.revokeRole, validator)

    def get_role_admin_changed_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for RoleAdminChanged event.

        :param tx_hash: hash of transaction emitting RoleAdminChanged event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=IPermissionsEnumerable.abi()).events.RoleAdminChanged().processReceipt(tx_receipt)
    def get_role_granted_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for RoleGranted event.

        :param tx_hash: hash of transaction emitting RoleGranted event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=IPermissionsEnumerable.abi()).events.RoleGranted().processReceipt(tx_receipt)
    def get_role_revoked_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for RoleRevoked event.

        :param tx_hash: hash of transaction emitting RoleRevoked event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=IPermissionsEnumerable.abi()).events.RoleRevoked().processReceipt(tx_receipt)

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"previousAdminRole","type":"bytes32"},{"indexed":true,"internalType":"bytes32","name":"newAdminRole","type":"bytes32"}],"name":"RoleAdminChanged","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleGranted","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"role","type":"bytes32"},{"indexed":true,"internalType":"address","name":"account","type":"address"},{"indexed":true,"internalType":"address","name":"sender","type":"address"}],"name":"RoleRevoked","type":"event"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleAdmin","outputs":[{"internalType":"bytes32","name":"","type":"bytes32"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"uint256","name":"index","type":"uint256"}],"name":"getRoleMember","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"}],"name":"getRoleMemberCount","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"grantRole","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"hasRole","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"renounceRole","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"role","type":"bytes32"},{"internalType":"address","name":"account","type":"address"}],"name":"revokeRole","outputs":[],"stateMutability":"nonpayable","type":"function"}]'  # noqa: E501 (line-too-long)
        )

# pylint: disable=too-many-lines



"""Generated wrapper for IERC1155 Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)

from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider

from zero_ex.contract_wrappers.bases import ContractMethod, Validator
from zero_ex.contract_wrappers.tx_params import TxParams


# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for IERC1155 below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        IERC1155Validator,
    )
except ImportError:

    class IERC1155Validator(  # type: ignore
        Validator
    ):
        """No-op input validator."""


try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass





class BalanceOfMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the balanceOf method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, owner: str, _id: int):
        """Validate the inputs to the balanceOf method."""
        self.validator.assert_valid(
            method_name='balanceOf',
            parameter_name='_owner',
            argument_value=owner,
        )
        owner = self.validate_and_checksum_address(owner)
        self.validator.assert_valid(
            method_name='balanceOf',
            parameter_name='_id',
            argument_value=_id,
        )
        # safeguard against fractional inputs
        _id = int(_id)
        return (owner, _id)

    def call(self, owner: str, _id: int, tx_params: Optional[TxParams] = None) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (owner, _id) = self.validate_and_normalize_inputs(owner, _id)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(owner, _id).call(tx_params.as_dict())
        return int(returned)

    def send_transaction(self, owner: str, _id: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (owner, _id) = self.validate_and_normalize_inputs(owner, _id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(owner, _id).transact(tx_params.as_dict())

    def build_transaction(self, owner: str, _id: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (owner, _id) = self.validate_and_normalize_inputs(owner, _id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(owner, _id).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, owner: str, _id: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (owner, _id) = self.validate_and_normalize_inputs(owner, _id)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(owner, _id).estimateGas(tx_params.as_dict())

class BalanceOfBatchMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the balanceOfBatch method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, owners: List[str], ids: List[int]):
        """Validate the inputs to the balanceOfBatch method."""
        self.validator.assert_valid(
            method_name='balanceOfBatch',
            parameter_name='_owners',
            argument_value=owners,
        )
        self.validator.assert_valid(
            method_name='balanceOfBatch',
            parameter_name='_ids',
            argument_value=ids,
        )
        return (owners, ids)

    def call(self, owners: List[str], ids: List[int], tx_params: Optional[TxParams] = None) -> List[int]:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (owners, ids) = self.validate_and_normalize_inputs(owners, ids)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(owners, ids).call(tx_params.as_dict())
        return [int(element) for element in returned]

    def send_transaction(self, owners: List[str], ids: List[int], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (owners, ids) = self.validate_and_normalize_inputs(owners, ids)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(owners, ids).transact(tx_params.as_dict())

    def build_transaction(self, owners: List[str], ids: List[int], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (owners, ids) = self.validate_and_normalize_inputs(owners, ids)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(owners, ids).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, owners: List[str], ids: List[int], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (owners, ids) = self.validate_and_normalize_inputs(owners, ids)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(owners, ids).estimateGas(tx_params.as_dict())

class IsApprovedForAllMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the isApprovedForAll method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, owner: str, operator: str):
        """Validate the inputs to the isApprovedForAll method."""
        self.validator.assert_valid(
            method_name='isApprovedForAll',
            parameter_name='_owner',
            argument_value=owner,
        )
        owner = self.validate_and_checksum_address(owner)
        self.validator.assert_valid(
            method_name='isApprovedForAll',
            parameter_name='_operator',
            argument_value=operator,
        )
        operator = self.validate_and_checksum_address(operator)
        return (owner, operator)

    def call(self, owner: str, operator: str, tx_params: Optional[TxParams] = None) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (owner, operator) = self.validate_and_normalize_inputs(owner, operator)
        tx_params = super().normalize_tx_params(tx_params)
        returned = self._underlying_method(owner, operator).call(tx_params.as_dict())
        return bool(returned)

    def send_transaction(self, owner: str, operator: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (owner, operator) = self.validate_and_normalize_inputs(owner, operator)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(owner, operator).transact(tx_params.as_dict())

    def build_transaction(self, owner: str, operator: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (owner, operator) = self.validate_and_normalize_inputs(owner, operator)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(owner, operator).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, owner: str, operator: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (owner, operator) = self.validate_and_normalize_inputs(owner, operator)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(owner, operator).estimateGas(tx_params.as_dict())

class SafeBatchTransferFromMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the safeBatchTransferFrom method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _from: str, to: str, ids: List[int], values: List[int], data: Union[bytes, str]):
        """Validate the inputs to the safeBatchTransferFrom method."""
        self.validator.assert_valid(
            method_name='safeBatchTransferFrom',
            parameter_name='_from',
            argument_value=_from,
        )
        _from = self.validate_and_checksum_address(_from)
        self.validator.assert_valid(
            method_name='safeBatchTransferFrom',
            parameter_name='_to',
            argument_value=to,
        )
        to = self.validate_and_checksum_address(to)
        self.validator.assert_valid(
            method_name='safeBatchTransferFrom',
            parameter_name='_ids',
            argument_value=ids,
        )
        self.validator.assert_valid(
            method_name='safeBatchTransferFrom',
            parameter_name='_values',
            argument_value=values,
        )
        self.validator.assert_valid(
            method_name='safeBatchTransferFrom',
            parameter_name='_data',
            argument_value=data,
        )
        return (_from, to, ids, values, data)

    def call(self, _from: str, to: str, ids: List[int], values: List[int], data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_from, to, ids, values, data) = self.validate_and_normalize_inputs(_from, to, ids, values, data)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(_from, to, ids, values, data).call(tx_params.as_dict())

    def send_transaction(self, _from: str, to: str, ids: List[int], values: List[int], data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (_from, to, ids, values, data) = self.validate_and_normalize_inputs(_from, to, ids, values, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, to, ids, values, data).transact(tx_params.as_dict())

    def build_transaction(self, _from: str, to: str, ids: List[int], values: List[int], data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (_from, to, ids, values, data) = self.validate_and_normalize_inputs(_from, to, ids, values, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, to, ids, values, data).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, _from: str, to: str, ids: List[int], values: List[int], data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (_from, to, ids, values, data) = self.validate_and_normalize_inputs(_from, to, ids, values, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, to, ids, values, data).estimateGas(tx_params.as_dict())

class SafeTransferFromMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the safeTransferFrom method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, _from: str, to: str, _id: int, value: int, data: Union[bytes, str]):
        """Validate the inputs to the safeTransferFrom method."""
        self.validator.assert_valid(
            method_name='safeTransferFrom',
            parameter_name='_from',
            argument_value=_from,
        )
        _from = self.validate_and_checksum_address(_from)
        self.validator.assert_valid(
            method_name='safeTransferFrom',
            parameter_name='_to',
            argument_value=to,
        )
        to = self.validate_and_checksum_address(to)
        self.validator.assert_valid(
            method_name='safeTransferFrom',
            parameter_name='_id',
            argument_value=_id,
        )
        # safeguard against fractional inputs
        _id = int(_id)
        self.validator.assert_valid(
            method_name='safeTransferFrom',
            parameter_name='_value',
            argument_value=value,
        )
        # safeguard against fractional inputs
        value = int(value)
        self.validator.assert_valid(
            method_name='safeTransferFrom',
            parameter_name='_data',
            argument_value=data,
        )
        return (_from, to, _id, value, data)

    def call(self, _from: str, to: str, _id: int, value: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (_from, to, _id, value, data) = self.validate_and_normalize_inputs(_from, to, _id, value, data)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(_from, to, _id, value, data).call(tx_params.as_dict())

    def send_transaction(self, _from: str, to: str, _id: int, value: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (_from, to, _id, value, data) = self.validate_and_normalize_inputs(_from, to, _id, value, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, to, _id, value, data).transact(tx_params.as_dict())

    def build_transaction(self, _from: str, to: str, _id: int, value: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (_from, to, _id, value, data) = self.validate_and_normalize_inputs(_from, to, _id, value, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, to, _id, value, data).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, _from: str, to: str, _id: int, value: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (_from, to, _id, value, data) = self.validate_and_normalize_inputs(_from, to, _id, value, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(_from, to, _id, value, data).estimateGas(tx_params.as_dict())

class SetApprovalForAllMethod(ContractMethod): # pylint: disable=invalid-name
    """Various interfaces to the setApprovalForAll method."""

    def __init__(self, web3_or_provider: Union[Web3, BaseProvider], contract_address: str, contract_function: ContractFunction, validator: Validator=None):
        """Persist instance data."""
        super().__init__(web3_or_provider, contract_address, validator)
        self._underlying_method = contract_function

    def validate_and_normalize_inputs(self, operator: str, approved: bool):
        """Validate the inputs to the setApprovalForAll method."""
        self.validator.assert_valid(
            method_name='setApprovalForAll',
            parameter_name='_operator',
            argument_value=operator,
        )
        operator = self.validate_and_checksum_address(operator)
        self.validator.assert_valid(
            method_name='setApprovalForAll',
            parameter_name='_approved',
            argument_value=approved,
        )
        return (operator, approved)

    def call(self, operator: str, approved: bool, tx_params: Optional[TxParams] = None) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        (operator, approved) = self.validate_and_normalize_inputs(operator, approved)
        tx_params = super().normalize_tx_params(tx_params)
        self._underlying_method(operator, approved).call(tx_params.as_dict())

    def send_transaction(self, operator: str, approved: bool, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (operator, approved) = self.validate_and_normalize_inputs(operator, approved)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(operator, approved).transact(tx_params.as_dict())

    def build_transaction(self, operator: str, approved: bool, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (operator, approved) = self.validate_and_normalize_inputs(operator, approved)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(operator, approved).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, operator: str, approved: bool, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (operator, approved) = self.validate_and_normalize_inputs(operator, approved)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(operator, approved).estimateGas(tx_params.as_dict())

# pylint: disable=too-many-public-methods,too-many-instance-attributes
class IERC1155:
    """Wrapper class for IERC1155 Solidity contract.

    All method parameters of type `bytes`:code: should be encoded as UTF-8,
    which can be accomplished via `str.encode("utf_8")`:code:.
    """
    balance_of: BalanceOfMethod
    """Constructor-initialized instance of
    :class:`BalanceOfMethod`.
    """

    balance_of_batch: BalanceOfBatchMethod
    """Constructor-initialized instance of
    :class:`BalanceOfBatchMethod`.
    """

    is_approved_for_all: IsApprovedForAllMethod
    """Constructor-initialized instance of
    :class:`IsApprovedForAllMethod`.
    """

    safe_batch_transfer_from: SafeBatchTransferFromMethod
    """Constructor-initialized instance of
    :class:`SafeBatchTransferFromMethod`.
    """

    safe_transfer_from: SafeTransferFromMethod
    """Constructor-initialized instance of
    :class:`SafeTransferFromMethod`.
    """

    set_approval_for_all: SetApprovalForAllMethod
    """Constructor-initialized instance of
    :class:`SetApprovalForAllMethod`.
    """


    def __init__(
        self,
        web3_or_provider: Union[Web3, BaseProvider],
        contract_address: str,
        validator: IERC1155Validator = None,
    ):
        """Get an instance of wrapper for smart contract.

        :param web3_or_provider: Either an instance of `web3.Web3`:code: or
            `web3.providers.base.BaseProvider`:code:
        :param contract_address: where the contract has been deployed
        :param validator: for validation of method inputs.
        """
        # pylint: disable=too-many-statements

        self.contract_address = contract_address

        if not validator:
            validator = IERC1155Validator(web3_or_provider, contract_address)

        web3 = None
        if isinstance(web3_or_provider, BaseProvider):
            web3 = Web3(web3_or_provider)
        elif isinstance(web3_or_provider, Web3):
            web3 = web3_or_provider
        else:
            raise TypeError(
                "Expected parameter 'web3_or_provider' to be an instance of either"
                + " Web3 or BaseProvider"
            )

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                         middleware['function'], layer=middleware['layer'],
                    )
            except ValueError as value_error:
                if value_error.args == ("You can't add the same un-named instance twice",):
                    pass

        self._web3_eth = web3.eth

        functions = self._web3_eth.contract(address=to_checksum_address(contract_address), abi=IERC1155.abi()).functions

        self.balance_of = BalanceOfMethod(web3_or_provider, contract_address, functions.balanceOf, validator)

        self.balance_of_batch = BalanceOfBatchMethod(web3_or_provider, contract_address, functions.balanceOfBatch, validator)

        self.is_approved_for_all = IsApprovedForAllMethod(web3_or_provider, contract_address, functions.isApprovedForAll, validator)

        self.safe_batch_transfer_from = SafeBatchTransferFromMethod(web3_or_provider, contract_address, functions.safeBatchTransferFrom, validator)

        self.safe_transfer_from = SafeTransferFromMethod(web3_or_provider, contract_address, functions.safeTransferFrom, validator)

        self.set_approval_for_all = SetApprovalForAllMethod(web3_or_provider, contract_address, functions.setApprovalForAll, validator)

    def get_approval_for_all_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for ApprovalForAll event.

        :param tx_hash: hash of transaction emitting ApprovalForAll event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=IERC1155.abi()).events.ApprovalForAll().processReceipt(tx_receipt)
    def get_transfer_batch_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for TransferBatch event.

        :param tx_hash: hash of transaction emitting TransferBatch event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=IERC1155.abi()).events.TransferBatch().processReceipt(tx_receipt)
    def get_transfer_single_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for TransferSingle event.

        :param tx_hash: hash of transaction emitting TransferSingle event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=IERC1155.abi()).events.TransferSingle().processReceipt(tx_receipt)
    def get_uri_event(
        self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """Get log entry for URI event.

        :param tx_hash: hash of transaction emitting URI event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=IERC1155.abi()).events.URI().processReceipt(tx_receipt)

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"_owner","type":"address"},{"indexed":true,"internalType":"address","name":"_operator","type":"address"},{"indexed":false,"internalType":"bool","name":"_approved","type":"bool"}],"name":"ApprovalForAll","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"_operator","type":"address"},{"indexed":true,"internalType":"address","name":"_from","type":"address"},{"indexed":true,"internalType":"address","name":"_to","type":"address"},{"indexed":false,"internalType":"uint256[]","name":"_ids","type":"uint256[]"},{"indexed":false,"internalType":"uint256[]","name":"_values","type":"uint256[]"}],"name":"TransferBatch","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"_operator","type":"address"},{"indexed":true,"internalType":"address","name":"_from","type":"address"},{"indexed":true,"internalType":"address","name":"_to","type":"address"},{"indexed":false,"internalType":"uint256","name":"_id","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"_value","type":"uint256"}],"name":"TransferSingle","type":"event"},{"anonymous":false,"inputs":[{"indexed":false,"internalType":"string","name":"_value","type":"string"},{"indexed":true,"internalType":"uint256","name":"_id","type":"uint256"}],"name":"URI","type":"event"},{"inputs":[{"internalType":"address","name":"_owner","type":"address"},{"internalType":"uint256","name":"_id","type":"uint256"}],"name":"balanceOf","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address[]","name":"_owners","type":"address[]"},{"internalType":"uint256[]","name":"_ids","type":"uint256[]"}],"name":"balanceOfBatch","outputs":[{"internalType":"uint256[]","name":"","type":"uint256[]"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"_owner","type":"address"},{"internalType":"address","name":"_operator","type":"address"}],"name":"isApprovedForAll","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"_from","type":"address"},{"internalType":"address","name":"_to","type":"address"},{"internalType":"uint256[]","name":"_ids","type":"uint256[]"},{"internalType":"uint256[]","name":"_values","type":"uint256[]"},{"internalType":"bytes","name":"_data","type":"bytes"}],"name":"safeBatchTransferFrom","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_from","type":"address"},{"internalType":"address","name":"_to","type":"address"},{"internalType":"uint256","name":"_id","type":"uint256"},{"internalType":"uint256","name":"_value","type":"uint256"},{"internalType":"bytes","name":"_data","type":"bytes"}],"name":"safeTransferFrom","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_operator","type":"address"},{"internalType":"bool","name":"_approved","type":"bool"}],"name":"setApprovalForAll","outputs":[],"stateMutability":"nonpayable","type":"function"}]'  # noqa: E501 (line-too-long)
        )

# pylint: disable=too-many-lines



from typing import Any, Dict, Optional

from web3 import Web3
from eth_account.account import LocalAccount
from thirdweb.constants.addresses import get_contract_address_by_chain_id
from thirdweb.constants.chains import ChainId
from thirdweb.core.classes.factory import ContractFactory
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.core.classes.provider_handler import ProviderHandler
from thirdweb.core.classes.registry import ContractRegistry
from thirdweb.types.contract import ContractType
from thirdweb.types.sdk import SDKOptions
from thirdweb.types.settings.metadata import (
    EditionDropContractMetadata,
    NFTDropContractMetadata,
    EditionContractMetadata,
    MarketplaceContractMetadata,
    NFTCollectionContractMetadata,
    TokenContractMetadata,
)


class ContractDeployer(ProviderHandler):
    __factory: ContractFactory
    __registry: ContractRegistry

    _storage: IpfsStorage

    def __init__(
        self,
        provider: Web3,
        signer: Optional[LocalAccount] = None,
        options: SDKOptions = SDKOptions(),
        storage: IpfsStorage = IpfsStorage(),
    ):
        super().__init__(provider, signer, options)
        self._storage = storage

    def deploy_nft_collection(self, metadata: NFTCollectionContractMetadata) -> str:
        """
        Deploy an NFT Collection contract.
        """

        return self._deploy_contract(ContractType.NFT_COLLECTION, metadata.to_json())

    def deploy_edition(self, metadata: EditionContractMetadata) -> str:
        """
        Deploy an Edition contract
        """

        return self._deploy_contract(ContractType.EDITION, metadata.to_json())

    def deploy_token(self, metadata: TokenContractMetadata) -> str:
        """
        Deploy a Token contract
        """

        return self._deploy_contract(ContractType.TOKEN, metadata.to_json())

    def deploy_marketplace(self, metadata: MarketplaceContractMetadata) -> str:
        """
        Deploy a Marketplace contract
        """

        return self._deploy_contract(ContractType.MARKETPLACE, metadata.to_json())

    def deploy_nft_drop(self, metadata: NFTDropContractMetadata) -> str:
        """
        Deploy an NFT Drop contract
        """

        return self._deploy_contract(ContractType.NFT_DROP, metadata.to_json())

    def deploy_edition_drop(self, metadata: EditionDropContractMetadata) -> str:
        """
        Deploy an Edition Drop contract
        """

        return self._deploy_contract(ContractType.EDITION_DROP, metadata.to_json())

    def _deploy_contract(
        self, contract_type: ContractType, contract_metadata: Dict[str, Any]
    ) -> str:
        factory = self._get_factory()
        return factory.deploy(contract_type, contract_metadata)

    def _get_registry(self) -> ContractRegistry:
        try:
            return self.__registry
        except:
            pass

        chain_id = ChainId(self.get_provider().eth.chain_id)
        registry_address = get_contract_address_by_chain_id(chain_id, "tw_registry")

        self.__registry = ContractRegistry(
            registry_address,
            self.get_provider(),
            self.get_signer(),
            self.get_options(),
        )

        return self.__registry

    def _get_factory(self) -> ContractFactory:
        try:
            return self.__factory
        except:
            pass

        chain_id = ChainId(self.get_provider().eth.chain_id)
        factory_address = get_contract_address_by_chain_id(chain_id, "tw_factory")

        self.__factory = ContractFactory(
            factory_address,
            self.get_provider(),
            self.get_signer(),
            self.get_options(),
            self._storage,
        )

        return self.__factory

    def _set_registry(self, registry: ContractRegistry):
        self.__registry = registry

    def _set_factory(self, factory: ContractFactory):
        self.__factory = factory
